from typing import Iterable
from src import symbols, wrappers
from src import types


def subscript(base: types.Expression, subscript: types.Expression) -> str:
    """Converts a subscript to LaTeX form."""
    base, subscript = types.params2expressions(base, subscript)
    return base + "_" + wrappers.wrap(subscript, "{", "}")


def superscript(base: types.Expression, superscript: types.Expression) -> str:
    """Converts a superscript to LaTeX form."""
    base, subscript = types.params2expressions(base, subscript)
    return base + "^" + wrappers.wrap(superscript, "{", "}")


def subscript_superscript(
    base: types.Expression, subscript: types.Expression, superscript: types.Expression
) -> str:
    """Converts a subscript and superscript to LaTeX form."""
    base, subscript, superscript = types.params2expressions(
        base, subscript, superscript
    )
    return superscript(subscript(base, subscript), superscript)


def arr2row(arr: Iterable) -> str:
    """Converts an array to LaTeX form."""
    return " & ".join(map(types.exp, arr))


def matrix(
    matrix: Iterable[Iterable], b: tuple[types.Bracket, types.Bracket] = None
) -> str:
    """Converts a matrix to LaTeX form."""
    matrix = wrappers.wrap_begin_end(r"\\".join(map(arr2row, matrix)), "matrix")
    return wrappers.brackets(matrix, (b[0], b[1])) if b is not None else matrix


def example_matrix(m: list[list], b: tuple[types.Bracket, types.Bracket] = None) -> str:
    """Creates a 4x4 matrix with dots to show the pattern."""
    cases = [0, symbols.dots("c"), symbols.dots("v"), symbols.dots("d")]
    if len(m) != 4 or len(m[0]) != 4:
        raise ValueError("Matrix must be 4x4.")
    for i in range(4):
        for j in range(4):
            cases[0] = types.exp(m[i][j])
            cond_case = int(f"{int(i==2)}{int(j==2)}", 2)
            m[i][j] = cases[cond_case]
    return matrix(m, b)


def fraction(numerator: types.Expression, denominator: types.Expression) -> str:
    """Converts a fraction to LaTeX form."""
    numerator, denominator = types.params2expressions(numerator, denominator)
    return f"\\frac" + "".join(wrappers.wrap(i, *r"{}") for i in [numerator, denominator])


def sum_latex(
    start: types.Expression, end: types.Expression, term: types.Expression
) -> str:
    """Converts a sum to LaTeX form."""
    start, end, term = types.params2expressions(start, end, term)
    return subscript_superscript("\\sum", start, end) + term


def product_latex(
    start: types.Expression, end: types.Expression, term: types.Expression
) -> str:
    """Converts a product to LaTeX form."""
    start, end, term = types.params2expressions(start, end, term)
    return subscript_superscript("\\prod", start, end) + term


def add_latex(*terms: types.Expression) -> str:
    """Formats terms into addition form."""
    terms = types.params2expressions(*terms)
    return " + ".join(terms)


def sub_latex(*terms: types.Expression) -> str:
    """Formats terms into subtraction form."""
    terms = types.params2expressions(*terms)
    return " - ".join(terms)


def mul_latex(
    *terms: types.Expression,
    operator: types.MulOperator = "",
) -> str:
    """Formats terms into multiplication form."""
    terms = types.params2expressions(*terms)
    op = symbols.mulop(operator)
    op = f" {op} " if op != "" else op
    return op.join(terms)


def div_latex(*terms: types.Expression) -> str:
    terms = types.params2expressions(*terms)
    return " \\div ".join(terms)


def sqrt_latex(base: types.Expression) -> str:
    """Converts a square root to LaTeX form."""
    base = types.exp(base)
    return r"\sqrt{" + base + r"}"


def nth_root_latex(base: types.Expression, n: types.Expression) -> str:
    """Converts a nth root to LaTeX form."""
    base, n = types.params2expressions(base, n)
    if n == "2":
        return sqrt_latex(base)
    return r"\sqrt" + wrappers.wrap(n, *"[]") + wrappers.wrap(base, *r"{}")


def log_latex(base: types.Expression, exp: types.Expression) -> str:
    """Converts a logarithm to LaTeX form."""
    base, exp = types.params2expressions(base, exp)
    return subscript("\\log", base) + exp


def lim_latex(base: types.Expression, exp: types.Expression) -> str:
    """Converts a limit to LaTeX form."""
    base, exp = types.params2expressions(base, exp)
    return subscript("\\lim", base) + exp


def integral_latex(
    start: types.Expression, end: types.Expression, term: types.Expression
) -> str:
    """Converts an integral to LaTeX form."""
    start, end, term = types.params2expressions(start, end, term)
    return subscript_superscript("\\int", start, end) + term



from typing import Literal
from .types import Bracket, Expression, exp
from .symbols import bracket


def wrap(expression: Expression, start, end) -> str:
    """Wraps an expression with start and end."""
    return start + exp(expression) + end


def wrap_begin_end(expression: Expression, mode: Literal["table", "matrix"]) -> str:
    """Wraps an expression with begin and end."""
    expression = exp(expression)
    return wrap(expression, f"\\begin{{{mode}}}", f"\\end{{{mode}}}")


def brackets(expression, b: tuple[Bracket, Bracket]) -> str:
    """Converts an expression to LaTeX form with brackets."""
    expression = exp(expression)
    l, r = map(bracket, b)
    return wrap(expression, f"\\left{l}", f"\\right{r}")



from src import matrix, example_matrix


def test_matrix():
    mat = [["x+y", 2], [3, 4]]
    assert matrix(mat).replace(" ", "") == r"\begin{matrix}x+y&2\\3&4\end{matrix}"


def test_matrix_brackets():
    mat = [["x+y", 2], [3, 4]]
    assert (
        matrix(mat, b=("(", ")")).replace(" ", "")
        == r"\left(\begin{matrix}x+y&2\\3&4\end{matrix}\right)"
    )


def test_example_matrix():
    mat = [
        ["n1", "n2", "n3", "n4"],
        ["n5", "n6", "n7", "n8"],
        ["n9", "n10", "n11", "n12"],
        ["n13", "n14", "n15", "n16"],
    ]
    assert (
        example_matrix(mat).replace(" ", "")
        == r"\begin{matrix}n1&n2&\cdots&n4\\n5&n6&\cdots&n8\\\vdots&\vdots&\ddots&\vdots\\n13&n14&\cdots&n16\end{matrix}"
    )



from typing import Literal
from .types import Bracket, MulOperator

NoEscapeBrackets = [
    "{",
    "}",
    "langle",
    "rangle",
    "lceil",
    "rceil",
    "lfloor",
    "rfloor",
]


def bracket(b: Bracket):
    return "\\" + b if b in NoEscapeBrackets else b


def mulop(op: MulOperator) -> str:
    if op not in [".", "x", ""]:
        raise ValueError("op must be either '.', 'x', or nothing")
    ops = {"x": "\\times", ".": "\\cdot"}
    return ops.get(op, op)


def dots(mode: Literal["l", "c", "v", "d"]) -> str:
    """Returns the appropriate dots for the mode."""
    if mode not in "lcvd":
        raise ValueError("mode must be either 'l', 'c', 'v', or 'd'")
    return f"\\{mode}dots"



from src.types import exp


def test_int():
    assert exp(1) == "1"


def test_float():
    assert exp(1.0) == "1.0"


def test_complex():
    assert exp(1 + 1j) == "1.0+i"


def test_str():
    assert exp("1") == "1"



from typing import Literal, Union

Bracket = Literal[
    "(",
    ")",
    "[",
    "]",
    "{",
    "}",
    "|",
    "\\|",
    "langle",
    "rangle",
    "lceil",
    "rceil",
    "lfloor",
    "rfloor",
]
MulOperator = Literal[".", "x", ""]


def __fmt_complex(c: complex) -> str:
    """Formats a complex number to LaTeX form."""
    if c.imag == 0:
        return str(c.real)
    if c.real == 0:
        return str(c.imag) + "i"
    if c.imag in [1, -1]:
        return str(c.real) + ("+" if c.imag == 1 else "-") + "i"
    return str(c).replace("(", "").replace(")", "").replace("j", "i")


MODES = {
    "str": lambda x: x,
    "int": lambda x: str(x),
    "float": lambda x: str(x),
    "complex": lambda x: __fmt_complex(x),
    "other": lambda x: None,
}
Expression = Union[str, int, float, complex]


def exp(expression: Expression) -> str:
    return MODES.get(type(expression).__name__, "other")(expression)


def params2expressions(*params: Expression) -> list[str]:
    return [exp(param) for param in params]



from src import mul_latex


def test_mul():
    assert mul_latex("x", "y") == "xy"
    assert mul_latex("x", "y", operator="x") == "x \\times y"
    assert mul_latex("x", "y", operator=".") == "x \\cdot y"



from setuptools import setup, find_packages

setup(
    name="math-tex-writer",
    version="0.0.1",
    description="Latex math expression writer",
    packages=find_packages(),
    python_requires=">=3.8",
    requires=[],
)
